#include <math.h>
#include <algorithm>

#include <float.h>

#include <QtCore/QFile>
#include <QtGui/QMouseEvent>
#include <QtGui/QMessageBox>
#include <QtGui/QImage>

#include "glwidget.h"

GLwidget::GLwidget(QWidget *parent, QGLFormat format)
: QGLWidget(format, parent)
, m_flightStepSize(0.01f)
, m_xTrans(0.0)
, m_yTrans(0.0)
, m_zTrans(-2)
, m_workVolumeSize(10)
, m_enableVFC(true)
, m_timer(0)
, m_enableFlight(false)
, m_renderedObjects(0)
, m_totalObjects(0)
{
	resetRotation();
	m_meshes.clear();
	m_meshStates.clear();
	srand( (unsigned)time( NULL ) );

	m_timer = new QTimer(this);
	connect(m_timer,SIGNAL(timeout()), this, SLOT(slotNextPathPos()));

	m_zTrans = -m_workVolumeSize;
}       

GLwidget::~GLwidget()
{
	cleanUpScene();
	cleanUpFlightPath();
}

void GLwidget::cleanUpScene()
{
	m_meshStates.clear();
	for( unsigned int i = 0; i < m_meshes.size(); i++)
		delete m_meshes[i];
	m_meshes.clear();
}

void GLwidget::cleanUpFlightPath()
{
	m_controlPoints.clear();
	m_flightPath.clear();
}

void GLwidget::initializeGL() 
{ 
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_RESCALE_NORMAL);
	glEnable(GL_CULL_FACE);
	glShadeModel(GL_SMOOTH); // we use flat shading
	glEnable(GL_LIGHTING); // enable lighting
	glEnable(GL_LIGHT0);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
	const char *version = (const char *)(glGetString(GL_VERSION));
	fprintf(stderr,"OpenGL Version %s",version);
}

void GLwidget::paintGL() 
{   
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);    

	glPushMatrix();
	glLoadIdentity();

	if ( m_enableFlight )
	{
		
		gluLookAt(m_fpCurPos->pos.x, m_fpCurPos->pos.y, m_fpCurPos->pos.z, 
				  m_fpCurPos->center.x, m_fpCurPos->center.y, m_fpCurPos->center.z, 
				  m_fpCurPos->up.x, m_fpCurPos->center.y, m_fpCurPos->center.z);        

	}
	else
	{
		glTranslatef(m_xTrans, m_yTrans, m_zTrans);
		glMultMatrixf(m_RotationMatrix);

		// draw flight path
		glDisable(GL_LIGHTING);
		glColor3f(0,1,0);
		glBegin(GL_LINE_STRIP);
		for( FlightPath::const_iterator cit = m_flightPath.begin(); cit != m_flightPath.end(); ++cit )
		{
			glVertex3f(cit->pos.x,cit->pos.y,cit->pos.z);
		}
		glEnd();

	}

	for( unsigned int i = 0; i < m_meshStates.size(); i++)
	{
		ObjectState object = m_meshStates[i];
		if ( object.visible || !m_enableVFC )
		{
			const Mesh *mesh = m_meshes[object.meshIdx];
			Vector3D center = ( mesh->getBBoxMax() + mesh->getBBoxMin() ) / 2;

			// *** UMBEDINGT BEACHTEN BEIM TEST MIT VIEWFRUSTUM ***
			glPushMatrix();

			glTranslatef( object.pos[0], object.pos[1], object.pos[2] );
			glScalef( object.scale, object.scale, object.scale );
			glTranslatef(-center[0], -center[1], -center[2]);
			mesh->draw();

			glPopMatrix();
		}
	}

	glPopMatrix();
}

void GLwidget::resizeGL( int width, int height ) 
{
	glViewport(0, 0, width, height);  
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();  

	double s = 0.1;
	m_left = -s;
	m_right = -m_left;
	m_bottom = -s * height/static_cast<double>(width);
	m_top = -m_bottom;
	m_Near = s;
	m_Far = m_workVolumeSize;
	glFrustum(m_left, m_right, m_bottom, m_top, m_Near, m_Far);
	//gluPerspective(60, r, 0.1, 10);

	glMatrixMode(GL_MODELVIEW);  
}

// set rotation Matrix for trackball to Identity
void GLwidget::resetRotation()
{
	for(int i = 1; i < 4; i++)
	{
		for(int j = 0; j < i; j++)
		{
			m_RotationMatrix[i*4+j] = m_RotationMatrix[j*4+i]= 0.0;
		}
	}
	m_RotationMatrix[0] = m_RotationMatrix[5] = m_RotationMatrix[10] = m_RotationMatrix[15] = 1.0;
}

void GLwidget::mousePressEvent( QMouseEvent * e /*the event*/ )
{
	m_lastPos = e->pos();
}

void GLwidget::mouseMoveEvent( QMouseEvent * e /*the event*/ )
{
	int dx = e->x() - m_lastPos.x();
	int dy = e->y() - m_lastPos.y();

	bool ctrl_key = e->modifiers() & Qt::MetaModifier;		// only needed for Mac OS X, but doesn't hurt on other OSes

	if ( (e->buttons() & Qt::RightButton) || ctrl_key ) // translation along z-Axis
	{
		m_zTrans += 0.2 * dy;
	}
	else if ( e->buttons() & Qt::MidButton ) // translation in xy-Plane
	{
		m_xTrans += 0.01 * dx;
		m_yTrans -= 0.01 * dy;
	}
	else if (e->buttons() & Qt::LeftButton)	// rotation
	{
		// openGL multiplies new transformations on the right
		// we want to apply rotation on the left
		glPushMatrix(); // push current openGL transform matrix to stack
		glLoadIdentity();
		glRotatef(0.5*dx,0.0, 1.0, 0.0);
		glRotatef(0.5*dy,1.0, 0.0, 0.0);
		glMultMatrixf(m_RotationMatrix);
		glGetFloatv(GL_MODELVIEW_MATRIX,m_RotationMatrix);
		glPopMatrix();
	}
	m_lastPos = e->pos();
	e->accept();
	updateGL();
}

bool GLwidget::slotLoadModels(const QString &filename) 
{
	cleanUpScene();

	FILE *file = fopen(filename.toAscii().data(),"r");
	unsigned int numMeshes = 0;
	unsigned int numInstances = 0;
	char cstr[256];
	unsigned int count;

	while ( !feof(file) )
	{
		if ( fscanf(file,"%s %d",cstr,&count) == 2 )
		{
			numMeshes++;
			numInstances += count;
		}
	}
	fclose(file);

	m_meshes.reserve(numMeshes);
	m_meshStates.reserve(numInstances);


	file = fopen(filename.toAscii().data(),"r");
	while ( !feof(file) )
	{
		if ( 2 != fscanf(file,"%s %d",cstr,&count) )
			continue;
		Mesh *mesh = new Mesh(cstr);
		if( !mesh->success() )
		{
			fprintf(stderr,"Failed reading obj file %s\n",cstr);;
		}
		else
		{
			Vector3D bboxSize( mesh->getBBoxMax() - mesh->getBBoxMin() );
			float unit_scale = 1.0f / std::max( bboxSize.x, std::max(bboxSize.y, bboxSize.z) );

			unsigned int meshIdx = m_meshes.size();
			for ( unsigned int i = 0; i < count; i++ )
			{
				float tx = getRandomNumber(-m_workVolumeSize/2,m_workVolumeSize/2);
				float ty = getRandomNumber(-m_workVolumeSize/2,m_workVolumeSize/2);
				float tz = getRandomNumber(-m_workVolumeSize/2,m_workVolumeSize/2);
				float scale = getRandomNumber(m_workVolumeSize/100,m_workVolumeSize/10);
				m_meshStates.push_back( ObjectState( true, meshIdx, Vector3D(tx,ty,tz), scale * unit_scale ) );
			}

			m_meshes.push_back(mesh);
		}
	}
	fclose(file);

	updateGL();
	return true;
}

void GLwidget::slotLoadConfiguration(const QString &s)
{
	m_meshStates.clear();

	FILE *file = fopen(s.toAscii().data(),"r");
	if ( 0 == file )
	{
		fprintf(stderr,"WARNING: Couldn't read file %s!\n",s.toAscii().data());
		return;
	}

	int numStates;
	int visible;
	unsigned int meshIdx;
	float tmp[3];
	float scale;
	fscanf(file,"%d\n",&numStates);
	m_meshStates.reserve(numStates);
	for( int i = 0; i < numStates; i++)
	{
		if ( 6 == fscanf(file,"%d %d %f %f %f %f\n",&visible, &meshIdx, &tmp[0], &tmp[1], &tmp[2], &scale ) )
		{
			m_meshStates.push_back( ObjectState( visible!=0, meshIdx, Vector3D(tmp), scale ) );
			if ( meshIdx >= m_meshes.size() )
			{
				fprintf(stderr,"WARNING: Scene Configuration file doesn't match Model file!\n Configuration file not loaded\n!");
				m_meshStates.clear();
				return;
			}
		}

	}

	fclose(file);

	updateGL();
}

void GLwidget::slotSaveConfiguration(const QString &s)
{
	FILE *file = fopen(s.toAscii().data(),"w");
	if ( 0 == file )
	{
		fprintf(stderr,"WARNING: Couldn't create file %s!\n",s.toAscii().data());
		return;
	}

	fprintf(file,"%d\n",m_meshStates.size());
	for( ObjectStateVec::const_iterator cit = m_meshStates.begin(); cit != m_meshStates.end(); ++cit)
		fprintf(file,"%d %d %f %f %f %f\n",cit->visible?1:0,cit->meshIdx,cit->pos.x,cit->pos.y,cit->pos.z,cit->scale);

	fclose(file);
}

void GLwidget::slotGenerateFlightPath()
{
	cleanUpFlightPath();
	// we generate 10 control points,
	// if you want to get more flexible, feel free to add a qt-widget in the GUI....
	const unsigned int numControlPoints = 10;

	Vector3D pos, viewdir, up;
	for( unsigned int i = 0; i < numControlPoints; i++)
	{
		for( int r = 0; r < 3; r++)
			pos[r] = getRandomNumber(-m_workVolumeSize/2,m_workVolumeSize/2);

		for( int r = 0; r < 3; r++ )
			viewdir[r] = getRandomNumber(-1,1);

		if ( viewdir.sqrLength() < FLT_EPSILON )
			viewdir = Vector3D(0,0,1);
		else
			viewdir.normalize();

		for( int r = 0; r < 3; r++)
			up[r] = getRandomNumber(-1,1);

		if ( up.sqrLength() < FLT_EPSILON )
			up = Vector3D(0,1,0);

		up = viewdir.crossProd(viewdir.crossProd(up));
		up.normalize();
		m_controlPoints.push_back( EyeState( pos, pos+viewdir, up ) );
	}

	computeFlightPath();

	emit signalFlightPathAvailable(m_flightPath.size()>0);

	updateGL();
}

void GLwidget::slotPlayFlightPath()
{
	if ( m_timer->isActive() )
	{
		m_timer->stop();
	}
	if ( m_flightPath.size() > 0 )
	{
		m_renderedObjects = 0;
		m_totalObjects = 0;
		emit signalRenderedObject(m_renderedObjects);
		emit signalTotalObjects(m_totalObjects);

		m_enableFlight = true;
		m_fpCurPos = m_flightPath.begin();
		updateGL();
		m_timer->start(40);
	}
}

void GLwidget::slotLoadFlightPath(const QString &s)
{
	cleanUpFlightPath();

	FILE *file = fopen(s.toAscii().data(),"r");
	if ( 0 == file )
	{
		fprintf(stderr,"WARNING: Couldn't read file %s!\n",s.toAscii().data());
		return;
	}
	float tmp[9];
	while ( !feof(file) )
	{
		if ( 9 == fscanf(file,"%f %f %f %f %f %f %f %f %f\n",&tmp[0],&tmp[1],&tmp[2],&tmp[3],&tmp[4],&tmp[5],&tmp[6],&tmp[7],&tmp[8]) )
			m_controlPoints.push_back( EyeState( Vector3D(&tmp[0]), Vector3D(&tmp[3]), Vector3D(&tmp[6]) ) );

	}
	fclose(file);

	computeFlightPath();

	emit signalFlightPathAvailable(m_flightPath.size()>0);

	updateGL();
}

void GLwidget::slotSaveFlightPath(const QString &s)
{
	FILE *file = fopen(s.toAscii().data(),"w");
	if ( 0 == file )
	{
		fprintf(stderr,"WARNING: Couldn't create file %s!\n",s.toAscii().data());
		return;
	}
	for ( FlightPath::const_iterator cit = m_controlPoints.begin(); cit != m_controlPoints.end(); ++cit )
	{
		fprintf(file,"%f %f %f %f %f %f %f %f %f\n",cit->pos.x,cit->pos.y,cit->pos.z,cit->center.x,cit->center.y,cit->center.z,cit->up.x,cit->up.y,cit->up.z);
	}
	fclose(file);
}

void GLwidget::slotNextPathPos()
{
	++m_fpCurPos;
	if ( m_fpCurPos == m_flightPath.end() )
	{
		m_timer->stop();
		m_fpCurPos = m_flightPath.begin();
		m_enableFlight = false;
	}

	computeObjectVisibility();

	for( unsigned int i = 0; i < m_meshStates.size(); i++)
		m_renderedObjects  += m_meshStates[i].visible ? 1 : 0;
	
	m_totalObjects += m_meshStates.size();

	emit signalRenderedObject(m_renderedObjects);
	emit signalTotalObjects(m_totalObjects);

	updateGL();
}

/*!
/* INPUT: m_controlPoints: at least 3 necessary
/* OUTPUT: m_flightPath
/* interpolates between control points through b-splines, samples the path and save resulting positions
*/
void GLwidget::computeFlightPath()
{
	if ( m_controlPoints.size() < 3)
	{
		fprintf(stderr,"WARNING: not enough control points: at least 4 needed!\n");
		return;
	}
	FlightPath::const_iterator cit = m_controlPoints.begin();
	EyeState p0, p1, p2;
	p0 = *cit; ++cit;
	p1 = *cit; ++cit;
	
	for ( ; cit != m_controlPoints.end(); ++cit)
	{
		p2 = *cit;
		for ( float t = 0; t < 1; t+=m_flightStepSize )
			m_flightPath.push_back( EyeState( evalQuadraticBspline(p0.pos,p1.pos,p2.pos,t),
											  evalQuadraticBspline(p0.center,p1.center,p2.center,t),	
											  evalQuadraticBspline(p0.up,p1.up,p2.up,t) ) );
		p0 = p1;
		p1 = p2;
	}
}

float GLwidget::evalQuadraticBspline(float P0, float P1, float P2, float t)
{
	float t2 = t*t;
	float b02 = t2 - 2*t + 1;
	float b12 = -2*t2 + 2*t + 1;
	float b22 = t2;
	return 0.5f * (b02*P0 + b12*P1 + b22*P2);
}

Vector3D GLwidget::evalQuadraticBspline(Vector3D P0, Vector3D P1, Vector3D P2, float t)
{
	float v[3];
	for( int d = 0; d < 3; d++ )
		v[d] = evalQuadraticBspline(P0[d],P1[d],P2[d],t);
	return Vector3D(v);
}

/*!
 * INPUTS: 	m_left, m_right, m_bottom, m_top, m_Near, m_Far    <-  View Frustum as set by glFrustum in resizeGL()
 *          m_flightPath  <- describes current camera viewpoint, direction, up-vector, as set by gluLookAt() in  paintGL()
 * OUTPUT: m_meshStates[i].visible for all object instances i
*/
void GLwidget::computeObjectVisibility()
{
	// ** TODO **

}
